import copy
import functools
import typing

import numpy as np
try:
    import pandas as pd
    PANDAS_INSTALLED = True
except ImportError:
    PANDAS_INSTALLED = False
from sklearn import base as sklearn_base
from sklearn import exceptions
from sklearn import pipeline
from sklearn import preprocessing
from sklearn import utils

from creme import base
from creme import compose
from creme import stream


__all__ = [
    'convert_creme_to_sklearn',
    'convert_sklearn_to_creme',
    'SKL2CremeClassifier',
    'SKL2CremeRegressor',
    'Creme2SKLRegressor',
    'Creme2SKLClassifier',
    'Creme2SKLClusterer',
    'Creme2SKLTransformer'
]

# Define a streaming method for each kind of batch input
STREAM_METHODS: typing.Dict[typing.Type, typing.Callable] = {
    np.ndarray: stream.iter_array
}

if PANDAS_INSTALLED:
    STREAM_METHODS[pd.DataFrame] = stream.iter_pandas

# Params passed to sklearn.utils.check_X_y and sklearn.utils.check_array
SKLEARN_INPUT_X_PARAMS = {
    'accept_sparse': False,
    'accept_large_sparse': True,
    'dtype': 'numeric',
    'order': None,
    'copy': False,
    'force_all_finite': True,
    'ensure_2d': True,
    'allow_nd': False,
    'ensure_min_samples': 1,
    'ensure_min_features': 1
}

# Params passed to sklearn.utils.check_X_y in addition to SKLEARN_INPUT_X_PARAMS
SKLEARN_INPUT_Y_PARAMS = {
    'multi_output': False,
    'y_numeric': False
}


def convert_creme_to_sklearn(estimator: base.Estimator):
    """Wraps a creme estimator to make it compatible with scikit-learn.

    Parameters:
        estimator

    """

    if isinstance(estimator, compose.Pipeline):
        return pipeline.Pipeline([
            (name, convert_creme_to_sklearn(step))
            for name, step in estimator.steps.items()
        ])

    wrappers = [
        (base.BinaryClassifier, Creme2SKLClassifier),
        (base.Clusterer, Creme2SKLClusterer),
        (base.MultiClassifier, Creme2SKLClassifier),
        (base.Regressor, Creme2SKLRegressor),
        (base.Transformer, Creme2SKLTransformer)
    ]

    for base_type, wrapper in wrappers:
        if isinstance(estimator, base_type):
            obj = wrapper(estimator)
            obj.instance_ = copy.deepcopy(estimator)
            return obj

    raise ValueError("Couldn't find an appropriate wrapper")


def convert_sklearn_to_creme(estimator: sklearn_base.BaseEstimator, batch_size=1,
                             classes: typing.Optional[list] = None):
    """Wraps a scikit-learn estimator to make it compatible with creme.

    Parameters:
        estimator
        batch_size: The amount of observations that will be used during each `partial_fit` call.
            Setting this to 1 means that the model will learn with each given observation.
            Increasing the batch size means that the observations will be stored in a buffer and
            the model will only update itself once enough observations are available.
        classes: Class names necessary for classifiers.

    """

    if not hasattr(estimator, 'partial_fit'):
        raise ValueError(f'{estimator} does not have a partial_fit method')

    wrappers = [
        (sklearn_base.RegressorMixin, functools.partial(
            SKL2CremeRegressor,
            batch_size=batch_size
        )),
        (sklearn_base.ClassifierMixin, functools.partial(
            SKL2CremeClassifier,
            batch_size=batch_size,
            classes=classes
        ))
    ]

    for base_type, wrapper in wrappers:
        if isinstance(estimator, base_type):
            return wrapper(copy.deepcopy(estimator))

    raise ValueError("Couldn't find an appropriate wrapper")


class SKL2CremeBase:

    def __init__(self, estimator: sklearn_base.BaseEstimator, batch_size: int,
                 x_dtype: typing.Type):
        self.estimator = estimator
        self.batch_size = batch_size
        self.x_dtype = x_dtype

        self._x_batch = None
        self._y_batch = None
        self._batch_i = 0


class SKL2CremeRegressor(SKL2CremeBase, base.Regressor):
    """`sklearn` to `creme` regressor adapter.

    Parameters:
        estimator: A scikit-learn transformer which has a `partial_fit` method.
        batch_size: The amount of observations that will be used during each `partial_fit` call.
            Setting this to 1 means that the model will learn with each given observation.
            Increasing the batch size means that the observations will be stored in a buffer and
            the model will only update itself once enough observations are available.

    Example:

        >>> from creme import compat
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import preprocessing
        >>> from creme import stream
        >>> from sklearn import linear_model
        >>> from sklearn import datasets

        >>> X_y = stream.iter_sklearn_dataset(
        ...     dataset=datasets.load_boston(),
        ...     shuffle=True,
        ...     seed=42
        ... )

        >>> scaler = preprocessing.StandardScaler()
        >>> sgd_reg = compat.convert_sklearn_to_creme(linear_model.SGDRegressor())
        >>> model = scaler | sgd_reg

        >>> metric = metrics.MAE()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MAE: 11.004415

    """

    def __init__(self, estimator: sklearn_base.RegressorMixin, batch_size=1):
        super().__init__(
            estimator=estimator,
            batch_size=batch_size,
            x_dtype=np.float
        )

    def fit_one(self, x, y):

        if self._x_batch is None:
            n_features = len(x)
            self._x_batch = np.empty(shape=(self.batch_size, n_features), dtype=self.x_dtype)
            self._y_batch = [None] * self.batch_size

        self._x_batch[self._batch_i, :] = list(x.values())
        self._y_batch[self._batch_i] = y

        self._batch_i += 1

        if self._batch_i == self.batch_size:
            self.estimator.partial_fit(X=self._x_batch, y=self._y_batch)
            self._batch_i = 0

        return self

    def predict_one(self, x):
        try:
            return self.estimator.predict(X=[list(x.values())])[0]
        except exceptions.NotFittedError:
            return 0


class SKL2CremeClassifier(SKL2CremeBase, base.MultiClassifier):
    """`sklearn` to `creme` classifier adapter.

    Parameters:
        estimator: A scikit-learn regressor which has a `partial_fit` method.
        classes
        batch_size: The amount of observations that will be used during each `partial_fit` call.
            Setting this to 1 means that the model will learn with each given observation.
            Increasing the batch size means that the observations will be stored in a buffer and
            the model will only update itself once enough observations are available.

    Example:

        >>> from creme import compat
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import preprocessing
        >>> from creme import stream
        >>> from sklearn import linear_model
        >>> from sklearn import datasets

        >>> X_y = stream.iter_sklearn_dataset(
        ...     dataset=datasets.load_breast_cancer(),
        ...     shuffle=True,
        ...     seed=42
        ... )

        >>> model = preprocessing.StandardScaler()
        >>> model |= compat.convert_sklearn_to_creme(
        ...     estimator=linear_model.SGDClassifier(
        ...         loss='log',
        ...         eta0=0.01,
        ...         learning_rate='constant'
        ...     ),
        ...     classes=[False, True]
        ... )

        >>> metric = metrics.LogLoss()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        LogLoss: 0.199554

    """

    def __init__(self, estimator: sklearn_base.ClassifierMixin, classes: list, batch_size=1):
        super().__init__(
            estimator=estimator,
            batch_size=batch_size,
            x_dtype=np.float
        )
        self.classes = classes

    def fit_one(self, x, y):

        if self._x_batch is None:
            n_features = len(x)
            self._x_batch = np.empty(shape=(self.batch_size, n_features), dtype=self.x_dtype)
            self._y_batch = [None] * self.batch_size

        self._x_batch[self._batch_i, :] = list(x.values())
        self._y_batch[self._batch_i] = y

        self._batch_i += 1

        if self._batch_i == self.batch_size:
            self.estimator.partial_fit(X=self._x_batch, y=self._y_batch, classes=self.classes)
            self._batch_i = 0

        return self

    def predict_proba_one(self, x):
        try:
            y_pred = self.estimator.predict_proba([list(x.values())])[0]
            return {self.classes[i]: p for i, p in enumerate(y_pred)}
        except exceptions.NotFittedError:
            return {c: 1 / len(self.classes) for c in self.classes}

    def predict_one(self, x):
        try:
            y_pred = self.estimator.predict(X=[list(x.values())])[0]
            return y_pred
        except exceptions.NotFittedError:
            return self.classes[0]


class Creme2SKLBase(sklearn_base.BaseEstimator):
    """This class is just here for house-keeping."""


class Creme2SKLRegressor(Creme2SKLBase, sklearn_base.RegressorMixin):
    """`creme` to `sklearn` regressor adapter.

    Parameters:
        estimator

    """

    def __init__(self, estimator: base.Regressor):
        self.estimator = estimator

    def fit(self, X, y):
        """Fits to an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))
            y (array-like of shape n_samples)

        Returns:
            self

        """

        # Check the estimator is a Regressor
        if not isinstance(self.estimator, base.Regressor):
            raise ValueError('estimator is not a Regressor')

        # Check the inputs
        X, y = utils.check_X_y(X, y, **SKLEARN_INPUT_X_PARAMS, **SKLEARN_INPUT_Y_PARAMS)

        # scikit-learn's convention is that fit shouldn't mutate the input parameters; we have to
        # deep copy the provided estimator in order to respect this convention
        self.instance_ = copy.deepcopy(self.estimator)

        # Call fit_one for each observation
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.instance_.fit_one(x, yi)

        return self

    def predict(self, X):
        """Predicts the target of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array of shape (n_samples,): Predicted target values for each row of `X`.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, attributes='instance_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        # Make a prediction for each observation
        y_pred = np.empty(shape=len(X))
        for i, (x, _) in enumerate(stream.iter_array(X)):
            y_pred[i] = self.instance_.predict_one(x)

        return y_pred


class Creme2SKLClassifier(Creme2SKLBase, sklearn_base.ClassifierMixin):
    """`creme` to `sklearn` classification adapter.

    Parameters:
        estimator

    """

    def __init__(self, estimator: base.Classifier):
        self.estimator = estimator

    def fit(self, X, y):
        """Fits to an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))
            y (array-like of shape n_samples)

        Returns:
            self

        """

        # Check the estimator is either a BinaryClassifier or a MultiClassifier
        if not isinstance(self.estimator, (base.BinaryClassifier, base.MultiClassifier)):
            raise ValueError('estimator is not a BinaryClassifier nor a MultiClassifier')

        # Check the inputs
        X, y = utils.check_X_y(X, y, **SKLEARN_INPUT_X_PARAMS, **SKLEARN_INPUT_Y_PARAMS)

        # Check the target
        utils.multiclass.check_classification_targets(y)

        # Check the number of classes agrees with the type of classifier
        self.classes_ = np.unique(y)
        if len(self.classes_) > 2 and not isinstance(self.estimator, base.MultiClassifier):
            raise ValueError(f'n_classes is more than 2 but {self.estimator} is a ' +
                             'BinaryClassifier')

        # creme's BinaryClassifier expects bools or 0/1 values
        if not isinstance(self.estimator, base.MultiClassifier):
            self.label_encoder_ = preprocessing.LabelEncoder().fit(y)
            y = self.label_encoder_.transform(y)

        # scikit-learn's convention is that fit shouldn't mutate the input parameters; we have to
        # deep copy the provided estimator in order to respect this convention
        self.instance_ = copy.deepcopy(self.estimator)

        # Call fit_one for each observation
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.instance_.fit_one(x, yi)

        # Store the number of features so that future inputs can be checked
        self.n_features_ = X.shape[1]

        return self

    def predict_proba(self, X):
        """Predicts the target probability of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array of shape (n_samples,): Predicted target values for each row of `X`.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, attributes='instance_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        if X.shape[1] != self.n_features_:
            raise ValueError(f'Expected {self.n_features_} features, got {X.shape[1]}')

        # creme's predictions have to converted to follow the scikit-learn conventions
        def reshape_probas(y_pred):
            return [y_pred.get(c, 0) for c in self.classes_]

        # Make a prediction for each observation
        y_pred = np.empty(shape=(len(X), len(self.classes_)))
        for i, (x, _) in enumerate(stream.iter_array(X)):
            y_pred[i] = reshape_probas(self.instance_.predict_proba_one(x))

        return y_pred

    def predict(self, X):
        """Predicts the target of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array of shape (n_samples,): Predicted target values for each row of `X`.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, attributes='instance_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        if X.shape[1] != self.n_features_:
            raise ValueError(f'Expected {self.n_features_} features, got {X.shape[1]}')

        # Make a prediction for each observation
        y_pred = [None] * len(X)
        for i, (x, _) in enumerate(stream.iter_array(X)):
            y_pred[i] = self.instance_.predict_one(x)

        return np.asarray(y_pred)


class Creme2SKLTransformer(Creme2SKLBase, sklearn_base.TransformerMixin):
    """`creme` to `sklearn` transformer adapter.

    Parameters:
        estimator

    """

    def __init__(self, estimator: base.Transformer):
        self.estimator = estimator

    def fit(self, X, y=None):
        """Fits to an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))
            y (array-like of shape n_samples)

        Returns:
            self

        """

        # Check the estimator is a Transformer
        if not isinstance(self.estimator, base.Transformer):
            raise ValueError('estimator is not a Transformer')

        # Check the inputs
        if y is None:
            X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)
        else:
            X, y = utils.check_X_y(X, y, **SKLEARN_INPUT_X_PARAMS, **SKLEARN_INPUT_Y_PARAMS)

        # scikit-learn's convention is that fit shouldn't mutate the input parameters; we have to
        # deep copy the provided estimator in order to respect this convention
        self.instance_ = copy.deepcopy(self.estimator)

        # Call fit_one for each observation
        if isinstance(self.instance_, base.SupervisedTransformer):
            for x, yi in STREAM_METHODS[type(X)](X, y):
                self.instance_.fit_one(x, yi)
        else:
            for x, _ in STREAM_METHODS[type(X)](X):
                self.instance_.fit_one(x)

        # Store the number of features so that future inputs can be checked
        self.n_features_ = X.shape[1]

        return self

    def transform(self, X):
        """Predicts the target of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array: Transformed output.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, attributes='instance_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        if X.shape[1] != self.n_features_:
            raise ValueError(f'Expected {self.n_features_} features, got {X.shape[1]}')

        # Call predict_proba_one for each observation
        X_trans = [None] * len(X)
        for i, (x, _) in enumerate(STREAM_METHODS[type(X)](X)):
            X_trans[i] = list(self.instance_.transform_one(x).values())

        return np.asarray(X_trans)


class Creme2SKLClusterer(Creme2SKLBase, sklearn_base.ClusterMixin):
    """Wraps a `creme` clusterer to make it compatible with `sklearn`.

    Parameters:
        estimator

    """

    def __init__(self, estimator: base.Clusterer):
        self.estimator = estimator

    def fit(self, X, y=None):
        """Fits to an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))
            y (array-like of shape n_samples)

        Returns:
            self

        """

        # Check the estimator is a Clusterer
        if not isinstance(self.estimator, base.Clusterer):
            raise ValueError('estimator is not a Clusterer')

        # Check the inputs
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        # scikit-learn's convention is that fit shouldn't mutate the input parameters; we have to
        # deep copy the provided estimator in order to respect this convention
        self.instance_ = copy.deepcopy(self.estimator)

        # Call fit_one for each observation
        self.labels_ = np.empty(len(X), dtype=np.int32)
        for i, (x, _) in enumerate(STREAM_METHODS[type(X)](X)):
            label = self.instance_.fit_one(x).predict_one(x)
            self.labels_[i] = label

        return self

    def predict(self, X):
        """Predicts the target of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array: Transformed output.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, attributes='instance_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        # Call predict_proba_one for each observation
        y_pred = np.empty(len(X), dtype=np.int32)
        for i, (x, _) in enumerate(STREAM_METHODS[type(X)](X)):
            y_pred[i] = self.instance_.predict_one(x)

        return y_pred



import collections
import copy
import functools

from creme import base


__all__ = ['OneVsOneClassifier']


class OneVsOneClassifier(base.Wrapper, base.MultiClassifier):
    """One-vs-One (OvO) multiclass strategy.

    This strategy consists in fitting one binary classifier for each pair of classes. Because we
    are in a streaming context, the number of classes isn't known from the start, hence new
    classifiers are instantiated on the fly.

    The number of classifiers is `k * (k - 1) / 2`, where `k` is the number of classes. However,
    each call to `fit_one` only requires training `k - 1` models. Indeed, only the models that
    pertain to the given label have to be trained. Meanwhile, making a prediction requires going
    through each and every model.

    Parameters:
        classifier: A binary classifier, although a multi-class classifier will work too.

    Attributes:
        classifiers (dict): A mapping between pairs of classes and classifiers. The keys are tuples
            which contain a pair of classes. Each pair is sorted in lexicographical order.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import multiclass
        >>> from creme import preprocessing

        >>> X_y = datasets.ImageSegments()

        >>> scaler = preprocessing.StandardScaler()
        >>> ovo = multiclass.OneVsOneClassifier(linear_model.LogisticRegression())
        >>> model = scaler | ovo

        >>> metric = metrics.MacroF1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MacroF1: 0.807573

    """

    def __init__(self, classifier):
        self.classifier = classifier
        new_clf = functools.partial(copy.deepcopy, classifier)
        self.classifiers = collections.defaultdict(new_clf)
        self.classes = set()

    @property
    def _wrapped_model(self):
        return self.classifier

    def fit_one(self, x, y):

        self.classes.add(y)

        for c in self.classes - {y}:
            pair = (c, y) if c < y else (y, c)
            self.classifiers[pair].fit_one(x, y=c < y)

        return self

    def predict_one(self, x):

        if not self.classifiers:  # is empty
            return None

        votes = collections.defaultdict(int)

        for pair, clf in self.classifiers.items():
            if clf.predict_one(x):
                votes[pair[1]] += 1
            else:
                votes[pair[0]] += 1

        return max(votes, key=votes.get)



import math

from creme import datasets
from creme import preprocessing
from creme import stream

import numpy as np
import pandas as pd


def test_standard_scaler_one_many_consistent():
    """Checks that using fit_one or fit_many produces the same result."""

    X = pd.read_csv(datasets.TrumpApproval().path)

    one = preprocessing.StandardScaler()
    for x, _ in stream.iter_pandas(X):
        one.fit_one(x)

    many = preprocessing.StandardScaler()
    for xb in np.array_split(X, 10):
        many.fit_many(xb)

    for i in X:
        assert math.isclose(one.counts[i], many.counts[i])
        assert math.isclose(one.means[i], many.means[i])
        assert math.isclose(one.vars[i], many.vars[i])


def test_standard_scaler_shuffle_columns():
    """Checks that fit_many works identically whether columns are shuffled or not."""

    X = pd.read_csv(datasets.TrumpApproval().path)

    normal = preprocessing.StandardScaler()
    for xb in np.array_split(X, 10):
        normal.fit_many(xb)

    shuffled = preprocessing.StandardScaler()
    for xb in np.array_split(X, 10):
        cols = np.random.permutation(X.columns)
        shuffled.fit_many(xb[cols])

    for i in X:
        assert math.isclose(shuffled.counts[i], shuffled.counts[i])
        assert math.isclose(shuffled.means[i], shuffled.means[i])
        assert math.isclose(shuffled.vars[i], shuffled.vars[i])


def test_standard_scaler_add_remove_columns():
    """Checks that no exceptions are raised whenever columns are dropped and/or added."""

    X = pd.read_csv(datasets.TrumpApproval().path)

    ss = preprocessing.StandardScaler()
    for xb in np.array_split(X, 10):
        # Pick half of the columns at random
        cols = np.random.choice(X.columns, len(X.columns) // 2, replace=False)
        ss.fit_many(xb[cols])



from . import base


__all__ = ['MAE']


class MAE(base.MeanMetric, base.RegressionMetric):
    """Mean absolute error.

    Example:

        >>> from creme import metrics

        >>> y_true = [3, -0.5, 2, 7]
        >>> y_pred = [2.5, 0.0, 2, 8]

        >>> metric = metrics.MAE()

        >>> for y_t, y_p in zip(y_true, y_pred):
        ...     print(metric.update(y_t, y_p).get())
        0.5
        0.5
        0.333
        0.5

        >>> metric
        MAE: 0.5

    """

    def _eval(self, y_true, y_pred):
        return abs(y_true - y_pred)



import collections

from . import base


__all__ = ['NesterovMomentum']


class NesterovMomentum(base.Optimizer):
    """Nesterov Momentum optimizer.

    Parameters:
        lr
        rho

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.NesterovMomentum()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.842932

    """

    def __init__(self, lr=.1, rho=.9):
        super().__init__(lr)
        self.rho = rho
        self.s = collections.defaultdict(float)

    def update_before_pred(self, w):

        for i in w:
            w[i] -= self.rho * self.s[i]

        return w

    def _update_after_pred(self, w, g):

        # Move w back to it's initial position
        for i in w:
            w[i] += self.rho * self.s[i]

        for i, gi in g.items():
            self.s[i] = self.rho * self.s[i] + self.learning_rate * gi
            w[i] -= self.s[i]

        return w



from .. import utils

from . import base
from . import quantile


class IQR(base.Univariate):
    """Computes the interquartile range.

    Parameters:
        q_inf: Desired inferior quantile, must be between 0 and 1. Defaults to `0.25`.
        q_sup: Desired superior quantile, must be between 0 and 1. Defaults to `0.75`.

    Example:

        >>> from creme import stats

        >>> iqr = stats.IQR(q_inf=0.25, q_sup=0.75)

        >>> for i in range(0, 1001):
        ...     iqr = iqr.update(i)
        ...     if i % 100 == 0:
        ...         print(iqr.get())
        0
        50.0
        100.0
        150.0
        200.0
        250.0
        300.0
        350.0
        400.0
        450.0
        500.0

    """

    def __init__(self, q_inf=.25, q_sup=.75):
        if q_inf >= q_sup:
            raise ValueError('q_inf must be strictly less than q_sup')
        self.q_inf = q_inf
        self.q_sup = q_sup
        self.quantile_inf = quantile.Quantile(q=self.q_inf)
        self.quantile_sup = quantile.Quantile(q=self.q_sup)

    @property
    def name(self):
        return f'{self.__class__.__name__}_{self.q_inf}_{self.q_sup}'

    def update(self, x):
        self.quantile_inf.update(x)
        self.quantile_sup.update(x)
        return self

    def get(self):
        return self.quantile_sup.get() - self.quantile_inf.get()


class RollingIQR(base.RollingUnivariate, utils.SortedWindow):
    """Computes the rolling interquartile range.

    Parameters:
        window_size (int): Size of the window.
        q_inf (float): Desired inferior quantile, must be between 0 and 1. Defaults to `0.25`.
        q_sup (float): Desired superior quantile, must be between 0 and 1. Defaults to `0.75`.

    Example:

        ::

            >>> from creme import stats
            >>> rolling_iqr = stats.RollingIQR(
            ...     q_inf=0.25,
            ...     q_sup=0.75,
            ...     window_size=100
            ... )

            >>> for i in range(0, 1001):
            ...     rolling_iqr = rolling_iqr.update(i)
            ...     if i % 100 == 0:
            ...         print(rolling_iqr.get())
            0
            50
            50
            50
            50
            50
            50
            50
            50
            50
            50

    """

    def __init__(self, window_size, q_inf=.25, q_sup=.75):
        super().__init__(size=window_size)
        if q_inf >= q_sup:
            raise ValueError('q_inf must be strictly less than q_sup')
        self.q_inf = q_inf
        self.q_sup = q_sup
        self.quantile_inf = quantile.RollingQuantile(q=self.q_inf, window_size=window_size)
        self.quantile_sup = quantile.RollingQuantile(q=self.q_sup, window_size=window_size)

    @property
    def name(self):
        return f'rolling_{self.__class__.__name__}_{self.q_inf}_{self.q_sup}'

    def update(self, x):
        self.quantile_inf.update(x)
        self.quantile_sup.update(x)
        return self

    def get(self):
        return self.quantile_sup.get() - self.quantile_inf.get()



import math
import collections

from . import base


__all__ = ['Nadam']


class Nadam(base.Optimizer):
    """Nadam optimizer.

    Parameters:
        lr
        beta_1
        beta_2
        eps

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.Nadam()
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.865961

    References:
        1. [Nadam: A combination of adam and nesterov](https://ruder.io/optimizing-gradient-descent/index.html#nadam)

    """

    def __init__(self, lr=.1, beta_1=.9, beta_2=.999, eps=1e-8):
        super().__init__(lr)
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.eps = eps
        self.m = collections.defaultdict(float)
        self.v = collections.defaultdict(float)

    def _update_after_pred(self, w, g):

        for i, gi in g.items():
            self.m[i] = self.beta_1 * self.m[i] + (1 - self.beta_1) * gi
            m_hat = self.m[i] / \
                (1 - math.pow(self.beta_1, self.n_iterations+1))

            self.v[i] = self.beta_2 * self.v[i] + (1 - self.beta_2) * gi ** 2
            v_hat = self.v[i] / \
                (1 - math.pow(self.beta_2, self.n_iterations+1))

            w[i] -= self.learning_rate * (self.beta_1 * m_hat + (1-self.beta_1)*gi/(
                1-math.pow(self.beta_1, self.n_iterations+1))) / (v_hat ** .5 + self.eps)

        return w



import collections
import copy
import functools
import importlib
import inspect
import math
import pickle
import random

from creme import metrics
from creme.metrics import base
import numpy as np
import pytest
from sklearn import metrics as sk_metrics


def load_metrics():
    """Yields all the metrics."""

    for name, obj in inspect.getmembers(importlib.import_module('creme.metrics'), inspect.isclass):

        if name == 'Metric':
            continue

        if issubclass(obj, metrics.Rolling):
            yield obj(metric=metrics.MSE(), window_size=42)
            continue

        elif name == 'RegressionMultiOutput':
            yield obj(metric=metrics.MSE())
            continue

        try:
            sig = inspect.signature(obj)
            yield obj(**{
                param.name: param.default if param.default != param.empty else 5
                for param in sig.parameters.values()
            })
        except ValueError:
            yield obj()


@pytest.mark.parametrize('metric', [
    pytest.param(
        metric,
        id=type(metric).__name__
    )
    for metric in load_metrics()
])
def test_pickling(metric):
    assert isinstance(pickle.loads(pickle.dumps(metric)), metric.__class__)
    assert isinstance(copy.deepcopy(metric), metric.__class__)


def generate_test_cases(metric, n):
    """Yields (y_true, y_pred) pairs of size n for a given metric."""

    sample_weights = [random.random() for _ in range(n)]

    if isinstance(metric, base.BinaryMetric):
        y_true = [random.choice([False, True]) for _ in range(n)]
        if metric.requires_labels:
            y_pred = [random.choice([False, True]) for _ in range(n)]
        else:
            y_pred = [np.random.dirichlet(np.ones(2)).tolist() for _ in range(n)]
        yield y_true, y_pred, sample_weights

    if isinstance(metric, base.MultiClassMetric):
        y_true = [random.choice([0, 1, 2]) for _ in range(n)]
        if metric.requires_labels:
            y_pred = [random.choice([0, 1, 2]) for _ in range(n)]
        else:
            y_pred = [np.random.dirichlet(np.ones(3)).tolist() for _ in range(n)]
        yield y_true, y_pred, sample_weights

    if isinstance(metric, base.RegressionMetric):
        yield (
            [random.random() for _ in range(n)],
            [random.random() for _ in range(n)],
            sample_weights
        )

def partial(f, **kwargs):
    return functools.update_wrapper(functools.partial(f, **kwargs), f)


TEST_CASES = [
    (metrics.Accuracy(), sk_metrics.accuracy_score),
    (metrics.Precision(), sk_metrics.precision_score),
    (metrics.MacroPrecision(), partial(sk_metrics.precision_score, average='macro')),
    (metrics.MicroPrecision(), partial(sk_metrics.precision_score, average='micro')),
    (metrics.WeightedPrecision(), partial(sk_metrics.precision_score, average='weighted')),
    (metrics.Recall(), sk_metrics.recall_score),
    (metrics.MacroRecall(), partial(sk_metrics.recall_score, average='macro')),
    (metrics.MicroRecall(), partial(sk_metrics.recall_score, average='micro')),
    (metrics.WeightedRecall(), partial(sk_metrics.recall_score, average='weighted')),
    (metrics.FBeta(beta=.5), partial(sk_metrics.fbeta_score, beta=.5)),
    (metrics.MacroFBeta(beta=.5), partial(sk_metrics.fbeta_score, beta=.5, average='macro')),
    (metrics.MicroFBeta(beta=.5), partial(sk_metrics.fbeta_score, beta=.5, average='micro')),
    (metrics.WeightedFBeta(beta=.5), partial(sk_metrics.fbeta_score, beta=.5, average='weighted')),
    (metrics.F1(), sk_metrics.f1_score),
    (metrics.MacroF1(), partial(sk_metrics.f1_score, average='macro')),
    (metrics.MicroF1(), partial(sk_metrics.f1_score, average='micro')),
    (metrics.WeightedF1(), partial(sk_metrics.f1_score, average='weighted')),
    (metrics.MCC(), sk_metrics.matthews_corrcoef),
    (metrics.MAE(), sk_metrics.mean_absolute_error),
    (metrics.MSE(), sk_metrics.mean_squared_error),
]


@pytest.mark.parametrize('metric, sk_metric', [
    pytest.param(
        metric,
        sk_metric,
        id=f'{metric.__class__.__name__}'
    )
    for metric, sk_metric in TEST_CASES
])
@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.filterwarnings('ignore::sklearn.metrics.classification.UndefinedMetricWarning')
def test_metric(metric, sk_metric):

    # Check str works
    str(metric)

    for y_true, y_pred, sample_weights in generate_test_cases(metric=metric, n=30):
        m = copy.deepcopy(metric)
        for i, (yt, yp, w) in enumerate(zip(y_true, y_pred, sample_weights)):

            if isinstance(yp, list):
                yp = dict(enumerate(yp))

            m.update(y_true=yt, y_pred=yp, sample_weight=w)

            if i >= 1:
                assert abs(
                    m.get() -
                    sk_metric(
                        y_true=y_true[:i + 1],
                        y_pred=y_pred[:i + 1],
                        sample_weight=sample_weights[:i + 1]
                    )
                ) < 1e-10


@pytest.mark.parametrize('metric, sk_metric', [
    pytest.param(
        metric,
        sk_metric,
        id=f'{metric.__class__.__name__}'
    )
    for metric, sk_metric in TEST_CASES
])
@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.filterwarnings('ignore::sklearn.metrics.classification.UndefinedMetricWarning')
def test_rolling_metric(metric, sk_metric):

    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    for n in (1, 2, 5, 10):
        for y_true, y_pred, _ in generate_test_cases(metric=metric, n=30):

            m = metrics.Rolling(metric=copy.deepcopy(metric), window_size=n)

            # Check str works
            str(m)

            for i, (yt, yp) in enumerate(zip(y_true, y_pred)):

                if isinstance(yp, list):
                    yp = dict(enumerate(yp))

                m.update(y_true=yt, y_pred=yp)

                if i >= 1:
                    assert abs(
                        m.get() -
                        sk_metric(
                            y_true=tail(y_true[:i + 1], n),
                            y_pred=tail(y_pred[:i + 1], n)
                        )
                    ) < 1e-10


def test_log_loss():

    metric = metrics.LogLoss()

    y_true = [True, False, False, True]
    y_pred = [0.9, 0.1, 0.2, 0.65]

    for i, (yt, yp) in enumerate(zip(y_true, y_pred)):
        metric.update(yt, yp)

        if i >= 1:
            assert math.isclose(metric.get(), sk_metrics.log_loss(y_true[:i + 1], y_pred[:i + 1]))

    metric.revert(y_true[-1], y_pred[-1])
    assert math.isclose(metric.get(), sk_metrics.log_loss(y_true[:-1], y_pred[:-1]))


def test_cross_entropy():

    metric = metrics.CrossEntropy()

    y_true = [0, 1, 2, 2]
    y_pred = [
        [0.29450637, 0.34216758, 0.36332605],
        [0.21290077, 0.32728332, 0.45981591],
        [0.42860913, 0.33380113, 0.23758974],
        [0.44941979, 0.32962558, 0.22095463]
    ]

    for i, (yt, yp) in enumerate(zip(y_true, y_pred)):
        yp = dict(enumerate(yp))
        metric.update(yt, yp)

        if i >= 1:
            assert math.isclose(
                metric.get(),
                sk_metrics.log_loss(y_true[:i + 1], y_pred[:i + 1], labels=[0, 1, 2])
            )

    metric.revert(y_true[-1], dict(enumerate(y_pred[-1])))
    assert math.isclose(metric.get(), sk_metrics.log_loss(y_true[:-1], y_pred[:-1]))


@pytest.mark.filterwarnings('ignore::sklearn.metrics.classification.UndefinedMetricWarning')
def test_multi_fbeta():

    fbeta = metrics.MultiFBeta(betas={0: 0.25, 1: 1, 2: 4}, weights={0: 1, 1: 1, 2: 2})
    sk_fbeta = sk_metrics.fbeta_score
    y_true = [0, 1, 2, 2, 2]
    y_pred = [0, 1, 0, 2, 1]

    for i, (yt, yp) in enumerate(zip(y_true, y_pred)):

        fbeta.update(yt, yp)

        if i >= 2:
            fbeta_0, _, _ = sk_fbeta(y_true[:i + 1], y_pred[:i + 1], beta=0.25, average=None)
            _, fbeta_1, _ = sk_fbeta(y_true[:i + 1], y_pred[:i + 1], beta=1, average=None)
            _, _, fbeta_2 = sk_fbeta(y_true[:i + 1], y_pred[:i + 1], beta=4, average=None)

            multi_fbeta = fbeta_0 * 1 + fbeta_1 * 1 + fbeta_2 * 2
            multi_fbeta /= (1 + 1 + 2)

            assert math.isclose(fbeta.get(), multi_fbeta)


@pytest.mark.filterwarnings('ignore::sklearn.metrics.classification.UndefinedMetricWarning')
def test_rolling_multi_f1():

    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    fbeta = metrics.Rolling(
        metric=metrics.MultiFBeta(
            betas={0: 0.25, 1: 1, 2: 4},
            weights={0: 1, 1: 1, 2: 2}
        ),
        window_size=3
    )
    n = fbeta.window_size
    sk_fbeta = sk_metrics.fbeta_score
    y_true = [0, 1, 2, 2, 2]
    y_pred = [0, 1, 0, 2, 1]

    for i, (yt, yp) in enumerate(zip(y_true, y_pred)):

        fbeta.update(yt, yp)

        if i >= 2:
            sk_y_true, sk_y_pred = tail(y_true[:i + 1], n), tail(y_pred[:i + 1], n)
            fbeta_0, _, _ = sk_fbeta(sk_y_true, sk_y_pred, beta=0.25, average=None)
            _, fbeta_1, _ = sk_fbeta(sk_y_true, sk_y_pred, beta=1, average=None)
            _, _, fbeta_2 = sk_fbeta(sk_y_true, sk_y_pred, beta=4, average=None)

            multi_fbeta = fbeta_0 * 1 + fbeta_1 * 1 + fbeta_2 * 2
            multi_fbeta /= (1 + 1 + 2)

            assert math.isclose(fbeta.get(), multi_fbeta)


def test_compose():

    metrics.MAE() + metrics.MSE()
    metrics.Accuracy() + metrics.LogLoss()
    metrics.Accuracy() + metrics.ConfusionMatrix()

    with pytest.raises(ValueError):
        _ = metrics.MSE() + metrics.LogLoss()

    with pytest.raises(ValueError):
        _ = metrics.MSE() + metrics.MAE() + metrics.LogLoss()



from creme import stream

from . import base


class Insects(base.RemoteDataset):
    """Insects dataset.

    This dataset has different variants, which are:

    - abrupt_balanced
    - abrupt_imbalanced
    - gradual_balanced
    - gradual_imbalanced
    - incremental-abrupt_balanced
    - incremental-abrupt_imbalanced
    - incremental-reoccurring_balanced
    - incremental-reoccurring_imbalanced
    - incremental_balanced
    - incremental_imbalanced
    - out-of-control

    The number of samples and the difficulty change from one variant to another. The number of
    classes is always the same (6), except for the last variant (24).

    Parameters:
        variant: Indicates which variant of the dataset to load.

    References:
        1. [USP DS repository](https://sites.google.com/view/uspdsrepository)
        2. [Souza, V., Reis, D.M.D., Maletzke, A.G. and Batista, G.E., 2020. Challenges in Benchmarking Stream Learning Algorithms with Real-world Data. arXiv preprint arXiv:2005.00113.](https://arxiv.org/abs/2005.00113)

    """

    variant_sizes = {
        'abrupt_balanced': (52848, 16419025),
        'abrupt_imbalanced': (355275, 110043637),
        'gradual_balanced': (24150, 7503750),
        'gradual_imbalanced': (143323, 44371501),
        'incremental-abrupt_balanced': (79986, 24849436),
        'incremental-abrupt_imbalanced': (452044, 140004225),
        'incremental-reoccurring_balanced': (79986, 24849092),
        'incremental-reoccurring_imbalanced': (452044, 140004230),
        'incremental_balanced': (57018, 17713574),
        'incremental_imbalanced': (452044, 140004218),
        'out-of-control': (905145, 277777854)
    }

    def __init__(self, variant: str):

        try:
            n_samples, size = self.variant_sizes[variant]
        except KeyError:
            variants = '\n'.join(f'- {v}' for v in self.variant_sizes)
            raise ValueError(f'Unknown variant, possible choices are:\n{variants}')
        n_classes = 24 if variant == 'out-of-control' else 6

        super().__init__(
            n_samples=n_samples,
            n_features=33,
            task=base.MULTI_CLF,
            url=f'http://sites.labic.icmc.usp.br/vsouza/repository/creme/INSECTS-{variant}_norm.arff',
            size=size,
            unpack=False
        )
        self.variant = variant

    def _iter(self):
        return stream.iter_arff(self.path, target='class')

    @property
    def _repr_title(self):
        return f'{self.__class__.__name__} dataset, {self.variant} variant'



from creme import compose
from creme import linear_model
from creme import preprocessing
import pytest


def test_pipeline_funcs():

    def a(x):
        pass

    def b(x):
        pass

    pipelines = [
        compose.FuncTransformer(a) | b,
        compose.FuncTransformer(a) | ('b', b),
        compose.FuncTransformer(a) | ('b', compose.FuncTransformer(b)),
        a | compose.FuncTransformer(b),
        ('a', a) | compose.FuncTransformer(b),
        ('a', compose.FuncTransformer(a)) | compose.FuncTransformer(b)
    ]

    for pipeline in pipelines:
        assert str(pipeline) == 'a | b'


def test_pipeline_add_at_start():

    def a(x):
        pass

    pipeline = preprocessing.StandardScaler() | linear_model.LinearRegression()
    pipeline = a | pipeline
    assert str(pipeline) == 'a | StandardScaler | LinearRegression'


def test_union_funcs():

    def a(x):
        pass

    def b(x):
        pass

    pipelines = [
        compose.FuncTransformer(a) + b,
        compose.FuncTransformer(a) + ('b', b),
        compose.FuncTransformer(a) + ('b', compose.FuncTransformer(b)),
        a + compose.FuncTransformer(b),
        ('a', a) + compose.FuncTransformer(b),
        ('a', compose.FuncTransformer(a)) + compose.FuncTransformer(b)
    ]

    for i, pipeline in enumerate(pipelines):
        print(i, str(pipeline))
        assert str(pipeline) == 'a + b'



import math

from creme import utils

from . import base


class Max(base.Univariate):
    """Running max.

    Attributes:
        max (float): The current max.

    Example:

        >>> from creme import stats

        >>> X = [1, -4, 3, -2, 5, -6]
        >>> _max = stats.Max()
        >>> for x in X:
        ...     print(_max.update(x).get())
        1
        1
        3
        3
        5
        5

    """

    def __init__(self):
        self.max = -math.inf

    def update(self, x):
        if x > self.max:
            self.max = x
        return self

    def get(self):
        return self.max


class RollingMax(base.RollingUnivariate, utils.SortedWindow):
    """Running max over a window.

    Parameters:
        window_size: Size of the rolling window.

    Example:

        >>> from creme import stats

        >>> X = [1, -4, 3, -2, 2, 1]
        >>> rolling_max = stats.RollingMax(window_size=2)
        >>> for x in X:
        ...     print(rolling_max.update(x).get())
        1
        1
        3
        3
        2
        2

    """

    def __init__(self, window_size: int):
        super().__init__(size=window_size)

    @property
    def window_size(self):
        return self.size

    def update(self, x):
        self.append(x)
        return self

    def get(self):
        return self[-1]


class AbsMax(base.Univariate):
    """Running absolute max.

    Attributes:
        abs_max (float): The current absolute max.

    Example:

        >>> from creme import stats

        >>> X = [1, -4, 3, -2, 5, -6]
        >>> abs_max = stats.AbsMax()
        >>> for x in X:
        ...     print(abs_max.update(x).get())
        1
        4
        4
        4
        5
        6

    """

    def __init__(self):
        self.abs_max = 0.

    def update(self, x):
        if abs(x) > self.abs_max:
            self.abs_max = abs(x)
        return self

    def get(self):
        return self.abs_max


class RollingAbsMax(base.RollingUnivariate, utils.SortedWindow):
    """Running absolute max over a window.

    Parameters:
        window_size: Size of the rolling window.

    Example:

        >>> from creme import stats

        >>> X = [1, -4, 3, -2, 2, 1]
        >>> rolling_absmax = stats.RollingAbsMax(window_size=2)
        >>> for x in X:
        ...     print(rolling_absmax.update(x).get())
        1
        4
        4
        3
        2
        2

    """

    def __init__(self, window_size: int):
        super().__init__(size=window_size)

    @property
    def window_size(self):
        return self.size

    def update(self, x):
        self.append(abs(x))
        return self

    def get(self):
        return self[-1]



import typing

import vaex
from vaex.utils import _ensure_strings_from_expressions, _ensure_list

from creme import base


def iter_vaex(X: vaex.dataframe.DataFrame, y: typing.Union[str, vaex.expression.Expression] = None,
              features: typing.Union[typing.List[str], vaex.expression.Expression] = None,
              **kwargs) -> base.typing.Stream:
    """Yields rows from a ``vaex.DataFrame``.

    Parameters:
        X: A vaex DataFrame housing the training featuers.
        y: The column or expression containing the target variable.
        features: A list of features used for training. If None, all columns in `X` will be used.
            Features specifying in `y` are ignored.

    """

    features = _ensure_strings_from_expressions(features)
    feature_names = features or X.get_column_names()

    if y:
        y = _ensure_strings_from_expressions(y)
        y = _ensure_list(y)
        feature_names = [feat for feat in feature_names if feat not in y]

    multioutput = len(y) > 1

    if multioutput:
        for i in range(len(X)):
            yield (
                {key: X.evaluate(key, i, i + 1)[0] for key in feature_names},
                {key: X.evaluate(key, i, i + 1)[0] for key in y}
            )

    else:

        for i in range(len(X)):
            yield (
                {key: X.evaluate(key, i, i + 1)[0] for key in feature_names},
                X.evaluate(y[0], i, i + 1)[0]
            )



import collections
import copy
import functools
import numbers
import typing

from creme import base
from creme import stats


__all__ = ['StatImputer']


class StatImputer(base.Transformer):
    """Replaces missing values with a statistic.

    This transformer allows you to replace missing values with the value of a running statistic.
    During a call to `fit_one`, for each feature, a statistic is updated whenever a numeric feature
    is observed. When `transform_one` is called, each feature with a `None` value is replaced with
    the current value of the corresponding statistic.

    Parameters:
        imputers: A list of tuples where each tuple has two elements. The first elements is a
            feature name and the second value is an instance of `stats.Univariate`. The second
            value can also be an arbitrary value, such as -1, in which case the missing values will
            be replaced with it.

    Example:

        >>> from creme import impute
        >>> from creme import stats

        For numeric data, we can use a `stats.Mean()` to replace missing values by the running
        average of the previously seen values:

        >>> X = [
        ...     {'temperature': 1},
        ...     {'temperature': 8},
        ...     {'temperature': 3},
        ...     {'temperature': None},
        ...     {'temperature': 4}
        ... ]

        >>> imp = impute.StatImputer(('temperature', stats.Mean()))

        >>> for x in X:
        ...     imp = imp.fit_one(x)
        ...     print(imp.transform_one(x))
        {'temperature': 1}
        {'temperature': 8}
        {'temperature': 3}
        {'temperature': 4.0}
        {'temperature': 4}

        For discrete/categorical data, a common practice is to `stats.Mode` to replace missing
        values by the most commonly seen value:

        >>> X = [
        ...     {'weather': 'sunny'},
        ...     {'weather': 'rainy'},
        ...     {'weather': 'sunny'},
        ...     {'weather': None},
        ...     {'weather': 'rainy'},
        ...     {'weather': 'rainy'},
        ...     {'weather': None}
        ... ]

        >>> imp = impute.StatImputer(('weather', stats.Mode()))

        >>> for x in X:
        ...     imp = imp.fit_one(x)
        ...     print(imp.transform_one(x))
        {'weather': 'sunny'}
        {'weather': 'rainy'}
        {'weather': 'sunny'}
        {'weather': 'sunny'}
        {'weather': 'rainy'}
        {'weather': 'rainy'}
        {'weather': 'rainy'}

        You can also choose to replace missing values with a constant value, as so:

        >>> imp = impute.StatImputer(('weather', 'missing'))

        >>> for x in X:
        ...     imp = imp.fit_one(x)
        ...     print(imp.transform_one(x))
        {'weather': 'sunny'}
        {'weather': 'rainy'}
        {'weather': 'sunny'}
        {'weather': 'missing'}
        {'weather': 'rainy'}
        {'weather': 'rainy'}
        {'weather': 'missing'}

        Multiple imputers can be defined by providing a tuple for each feature which you want to
        impute:

        >>> X = [
        ...     {'weather': 'sunny', 'temperature': 8},
        ...     {'weather': 'rainy', 'temperature': 3},
        ...     {'weather': 'sunny', 'temperature': None},
        ...     {'weather': None, 'temperature': 4},
        ...     {'weather': 'snowy', 'temperature': -4},
        ...     {'weather': 'snowy', 'temperature': -3},
        ...     {'weather': 'snowy', 'temperature': -3},
        ...     {'weather': None, 'temperature': None}
        ... ]

        >>> imp = impute.StatImputer(
        ...     ('temperature', stats.Mean()),
        ...     ('weather', stats.Mode())
        ... )

        >>> for x in X:
        ...     imp = imp.fit_one(x)
        ...     print(imp.transform_one(x))
        {'weather': 'sunny', 'temperature': 8}
        {'weather': 'rainy', 'temperature': 3}
        {'weather': 'sunny', 'temperature': 5.5}
        {'weather': 'sunny', 'temperature': 4}
        {'weather': 'snowy', 'temperature': -4}
        {'weather': 'snowy', 'temperature': -3}
        {'weather': 'snowy', 'temperature': -3}
        {'weather': 'snowy', 'temperature': 0.8333}

        A sophisticated way to go about imputation is condition the statistics on a given feature.
        For instance, we might want to replace a missing temperature with the average temperature
        of a particular weather condition. As an example, consider the following dataset where the
        temperature is missing, but not the weather condition:

        >>> X = [
        ...     {'weather': 'sunny', 'temperature': 8},
        ...     {'weather': 'rainy', 'temperature': 3},
        ...     {'weather': 'sunny', 'temperature': None},
        ...     {'weather': 'rainy', 'temperature': 4},
        ...     {'weather': 'sunny', 'temperature': 10},
        ...     {'weather': 'sunny', 'temperature': None},
        ...     {'weather': 'sunny', 'temperature': 12},
        ...     {'weather': 'rainy', 'temperature': None}
        ... ]

        Each missing temperature can be replaced with the average temperature of the corresponding
        weather condition as so:

        >>> from creme import compose

        >>> imp = compose.Grouper(
        ...     impute.StatImputer(('temperature', stats.Mean())),
        ...     by='weather'
        ... )

        >>> for x in X:
        ...     imp = imp.fit_one(x)
        ...     print(imp.transform_one(x))
        {'weather': 'sunny', 'temperature': 8}
        {'weather': 'rainy', 'temperature': 3}
        {'weather': 'sunny', 'temperature': 8.0}
        {'weather': 'rainy', 'temperature': 4}
        {'weather': 'sunny', 'temperature': 10}
        {'weather': 'sunny', 'temperature': 9.0}
        {'weather': 'sunny', 'temperature': 12}
        {'weather': 'rainy', 'temperature': 3.5}

        Note that you can also create a `Grouper` with the `*` operator:

        >>> imp = impute.StatImputer(('temperature', stats.Mean())) * 'weather'

    """

    def __init__(self, *imputers):
        self.stats = {
            feature: stat if isinstance(stat, stats.Univariate) else Constant(stat)
            for feature, stat in imputers
        }

    def fit_one(self, x):

        for i in self.stats:
            if x[i] is not None:
                self.stats[i].update(x[i])

        return self

    def transform_one(self, x):

        # Transformers are supposed to be pure, therefore we make a copy of the features
        x = x.copy()

        for i in self.stats:
            if x[i] is None:
                x[i] = self.stats[i].get()

        return x


class Constant(stats.Univariate):
    """Implements the `stats.Univariate` interface but always returns the same value.

    Parameters:
        value

    """

    def __init__(self, value: typing.Any):
        self.value = value

    def update(self, x):
        return self

    def get(self):
        return self.value

    @property
    def name(self):
        return self.value



"""Ensemble learning."""
from .bagging import BaggingClassifier
from .bagging import BaggingRegressor
from .boosting import AdaBoostClassifier
from .hedging import HedgeRegressor
from .stacking import StackingBinaryClassifier


__all__ = [
    'AdaBoostClassifier',
    'BaggingClassifier',
    'BaggingRegressor',
    'HedgeRegressor',
    'StackingBinaryClassifier',
]



import typing

from creme import base


__all__ = ['FuncTransformer']


class FuncTransformer(base.Transformer):
    """Wraps a function to make it usable in a pipeline.

    There is often a need to apply an arbitrary transformation to a set of features. For instance,
    this could involve parsing a date and then extracting the hour from said date. If you're
    processing a stream of data, then you can do this yourself by calling the necessary code at
    your leisure. On the other hand, if you want to do this as part of a pipeline, then you need to
    follow a simple convention.

    To use a function as part of a pipeline, take as input a `dict` of features and output a `dict`.
    Once you have initialized this class with your function, then you can use it like you would use
    any other (unsupervised) transformer.

    It is up to you if you want your function to be pure or not. By pure we refer to a function
    that doesn't modify its input. However, we recommend writing pure functions because this
    reduces the chances of inserting bugs into your pipeline.

    Parameters:
        func: A function that takes as input a `dict` and outputs a `dict`.

    Example:

        >>> from pprint import pprint
        >>> import datetime as dt
        >>> from creme import compose

        >>> x = {'date': '2019-02-14'}

        >>> def parse_date(x):
        ...     date = dt.datetime.strptime(x['date'], '%Y-%m-%d')
        ...     x['is_weekend'] = date.day in (5, 6)
        ...     x['hour'] = date.hour
        ...     return x

        >>> t = compose.FuncTransformer(parse_date)
        >>> pprint(t.transform_one(x))
        {'date': '2019-02-14', 'hour': 0, 'is_weekend': False}

        The above example is not pure because it modifies the input. The following example is pure
        and produces the same output:

        >>> def parse_date(x):
        ...     date = dt.datetime.strptime(x['date'], '%Y-%m-%d')
        ...     return {'is_weekend': date.day in (5, 6), 'hour': date.hour}

        >>> t = compose.FuncTransformer(parse_date)
        >>> pprint(t.transform_one(x))
        {'hour': 0, 'is_weekend': False}

        The previous example doesn't include the `date` feature because it returns a new `dict`.
        However, a common usecase is to add a feature to an existing set of features. You can do
        this in a pure way by unpacking the input `dict` into the output `dict`:

        >>> def parse_date(x):
        ...     date = dt.datetime.strptime(x['date'], '%Y-%m-%d')
        ...     return {'is_weekend': date.day in (5, 6), 'hour': date.hour, **x}

        >>> t = compose.FuncTransformer(parse_date)
        >>> pprint(t.transform_one(x))
        {'date': '2019-02-14', 'hour': 0, 'is_weekend': False}

        You can add `FuncTransformer` to a pipeline just like you would with any other transformer.

        >>> from creme import naive_bayes

        >>> pipeline = compose.FuncTransformer(parse_date) | naive_bayes.MultinomialNB()
        >>> pipeline
        Pipeline (
          FuncTransformer (
            func="parse_date"
          ),
          MultinomialNB (
            alpha=1.
          )
        )

        If you provide a function with wrapping it, then the pipeline will do it for you:

        >>> pipeline = parse_date | naive_bayes.MultinomialNB()

    """

    def __init__(self, func: typing.Callable[[dict], dict]):
        self.func = func

    def transform_one(self, x):
        return self.func(x)

    def __str__(self):
        return self.func.__name__



from . import base


__all__ = ['SMAPE']


class SMAPE(base.MeanMetric, base.RegressionMetric):
    """Symmetric mean absolute percentage error.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 0.07533, 0.07533, 0.07533, 0.07533, 0.07533, 0.07533, 0.0672, 0.0672]
        >>> y_pred = [0, 0.102, 0.107, 0.047, 0.1, 0.032, 0.047, 0.108, 0.089]

        >>> metric = metrics.SMAPE()
        >>> for yt, yp in zip(y_true, y_pred):
        ...     metric = metric.update(yt, yp)

        >>> metric
        SMAPE: 37.869392

    """

    def _eval(self, y_true, y_pred):
        den = abs(y_true) + abs(y_pred)
        if den == 0:
            return 0.
        return 2. * abs(y_true - y_pred) / den

    def get(self):
        return 100 * super().get()



import collections
import copy

from creme import base


__all__ = ['ClassifierChain', 'RegressorChain']


class BaseChain(base.Wrapper, collections.UserDict):

    def __init__(self, model, order=None):
        super().__init__()
        self.model = model
        self.order = order

        # If the order is specified, then we can instantiate a model for each label, if not we'll
        # do it in the first call to fit_one
        if order is not None:
            self._init_models()

    @property
    def _wrapped_model(self):
        return self.model

    def _init_models(self):
        for o in self.order:
            self[o] = copy.deepcopy(self.model)


class ClassifierChain(BaseChain, base.MultiOutputClassifier):
    """A multi-output model that arranges classifiers into a chain.

    This will create one model per output. The prediction of the first output will be used as a
    feature in the second output. The prediction for the second output will be used as a feature
    for the third, etc. This "chain model" is therefore capable of capturing dependencies between
    outputs.

    Parameters:
        model
        order: The order in which to construct the chain. If it not provided then it will be
            inferred from the order of the keys in the first provided target dictionary.

    Example:

        >>> from creme import feature_selection
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import multioutput
        >>> from creme import preprocessing
        >>> from creme import stream
        >>> from sklearn import datasets

        >>> X_y = stream.iter_sklearn_dataset(
        ...     dataset=datasets.fetch_openml('yeast', version=4),
        ...     shuffle=True,
        ...     seed=42
        ... )

        >>> model = feature_selection.VarianceThreshold(threshold=0.01)
        >>> model |= preprocessing.StandardScaler()
        >>> model |= multioutput.ClassifierChain(
        ...     model=linear_model.LogisticRegression(),
        ...     order=list(range(14))
        ... )

        >>> metric = metrics.Jaccard()

        >>> for x, y in X_y:
        ...     # Convert y values to booleans
        ...     y = {i: yi == 'TRUE' for i, yi in y.items()}
        ...     y_pred = model.predict_one(x)
        ...     metric = metric.update(y, y_pred)
        ...     model = model.fit_one(x, y)

        >>> metric
        Jaccard: 0.451524

    References:
        1. [Multi-Output Chain Models and their Application in Data Streams](https://jmread.github.io/talks/2019_03_08-Imperial_Stats_Seminar.pdf)

    """

    def __init__(self, model: base.Classifier, order: list = None):
        super().__init__(model, order)

    def fit_one(self, x, y):

        if self.order is None:
            self.order = list(y.keys())
            self._init_models()

        x = copy.copy(x)

        for o in self.order:
            clf = self[o]

            # Make predictions before the model is updated to avoid leakage
            y_pred = clf.predict_proba_one(x)

            clf.fit_one(x, y[o])

            # The predictions are stored as features for the next label
            if isinstance(clf, base.BinaryClassifier):
                x[o] = y_pred[True]
            else:
                for label, proba in y_pred.items():
                    x[f'{o}_{label}'] = proba

        return self

    def predict_proba_one(self, x):

        x = copy.copy(x)
        y_pred = {}

        for o in self.order:
            clf = self[o]

            y_pred[o] = clf.predict_proba_one(x)

            # The predictions are stored as features for the next label
            if isinstance(clf, base.BinaryClassifier):
                x[o] = y_pred[o][True]
            else:
                for label, proba in y_pred.items():
                    x[f'{o}_{label}'] = proba

        return y_pred


class RegressorChain(BaseChain, base.MultiOutputRegressor):
    """A multi-output model that arranges regressor into a chain.

    This will create one model per output. The prediction of the first output will be used as a
    feature in the second output. The prediction for the second output will be used as a feature
    for the third, etc. This "chain model" is therefore capable of capturing dependencies between
    outputs.

    Parameters:
        model
        order: The order in which to construct the chain. If it not provided then it will be
            inferred from the order of the keys in the first provided target dictionary.

    Example:

        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import multioutput
        >>> from creme import preprocessing
        >>> from creme import stream
        >>> from sklearn import datasets

        >>> X_y = stream.iter_sklearn_dataset(
        ...     dataset=datasets.load_linnerud(),
        ...     shuffle=True,
        ...     seed=42
        ... )

        >>> model = multioutput.RegressorChain(
        ...     model=(
        ...         preprocessing.StandardScaler() |
        ...         linear_model.LinearRegression(intercept_lr=0.3)
        ...     ),
        ...     order=[0, 1, 2]
        ... )

        >>> metric = metrics.RegressionMultiOutput(metrics.MAE())

        >>> model_selection.progressive_val_score(X_y, model, metric)
        MAE: 16.396347

    """

    def __init__(self, model: base.Regressor, order: list = None):
        super().__init__(model, order)

    def fit_one(self, x, y):

        if self.order is None:
            self.order = list(y.keys())
            self._init_models()

        x = copy.copy(x)

        for o in self.order:
            reg = self[o]

            # Make predictions before the model is updated to avoid leakage
            y_pred = reg.predict_one(x)

            reg.fit_one(x, y[o])

            # The predictions are stored as features for the next label
            x[o] = y_pred

        return self

    def predict_one(self, x):

        x = copy.copy(x)
        y_pred = {}

        for o, clf in self.items():
            y_pred[o] = clf.predict_one(x)
            x[o] = y_pred[o]

        return y_pred



import copy
import functools
import itertools
import math
import random

import numpy as np
import pandas as pd
import pytest

from creme import datasets
from creme import linear_model
from creme import optim
from creme import preprocessing
from creme import stream
from creme import utils


def iter_perturbations(keys, n=10):
    """Enumerate pertubations that will be applied to the weights."""

    # Enumerate unit vectors
    for i in keys:
        yield {j: int(i == j) for j in keys}

    # Enumerate samples from a spherical Gaussian
    for _ in range(n):
        p = {j: random.gauss(0, 1) for j in keys}
        norm = utils.math.norm(p, order=2)
        for j in p:
            p[j] /= norm
        yield p


@pytest.mark.parametrize(
    'lm, X_y',
    [
        pytest.param(
            lm(optimizer=copy.deepcopy(optimizer), initializer=initializer, l2=0),
            X_y(),
            id=f'{lm.__name__} - {optimizer} - {initializer}'
        )
        for lm, X_y in [
            (linear_model.LinearRegression, datasets.TrumpApproval),
            (linear_model.LogisticRegression, datasets.Bananas)
        ]
        for optimizer, initializer in itertools.product(
            [
                optim.AdaBound(),
                optim.AdaDelta(),
                optim.AdaGrad(),
                optim.AdaMax(),
                optim.Adam(),
                optim.AMSGrad(),
                # TODO: check momentum optimizers
                #optim.Momentum(),
                #optim.NesterovMomentum(),
                optim.RMSProp(),
                optim.SGD()
            ],
            [
                optim.initializers.Zeros(),
                optim.initializers.Normal(mu=0, sigma=1, seed=42)
            ]
        )
    ]
)
@pytest.mark.slow
def test_finite_differences(lm, X_y):
    """Checks the gradient of a linear model via finite differences.

    References:
        1. [How to test gradient implementations](https://timvieira.github.io/blog/post/2017/04/21/how-to-test-gradient-implementations/)
        2. [Stochastic Gradient Descent Tricks](https://cilvr.cs.nyu.edu/diglib/lsml/bottou-sgd-tricks-2012.pdf)

    """

    scaler = preprocessing.StandardScaler()
    eps = 1e-6

    for x, y in X_y:

        print(x)
        print('---------')
        x = scaler.fit_one(x).transform_one(x)
        print(x)
        print('+++++++++')

        # Store the current gradient and weights
        gradient, _ = lm._eval_gradient_one(x, y, 1)
        weights = lm.weights.copy()

        # d is a set of weight perturbations
        for d in iter_perturbations(weights.keys()):

            # Pertubate the weights and obtain the loss with the new weights
            lm.weights = {i: weights[i] + eps * di for i, di in d.items()}
            forward = lm.loss(y_true=y, y_pred=lm._raw_dot_one(x))
            lm.weights = {i: weights[i] - eps * di for i, di in d.items()}
            backward = lm.loss(y_true=y, y_pred=lm._raw_dot_one(x))

            # We expect g and h to be equal
            g = utils.math.dot(d, gradient)
            h = (forward - backward) / (2 * eps)

            # Compare signs
            # TODO: reactivate this check
            #assert np.sign(g) == np.sign(h)

            # Check absolute difference
            # TODO: decrease the tolerance
            assert abs(g - h) < 1e-5

        # Reset the weights to their original values in order not to influence
        # the training loop, even though it doesn't really matter.
        lm.weights = weights
        lm.fit_one(x, y)


def test_one_many_consistent():
    """Checks that using fit_one or fit_many produces the same result."""

    X = pd.read_csv(datasets.TrumpApproval().path)
    Y = X.pop('five_thirty_eight')

    one = linear_model.LinearRegression()
    for x, y in stream.iter_pandas(X, Y):
        one.fit_one(x, y)

    many = linear_model.LinearRegression()
    for xb, yb in zip(np.array_split(X, len(X)), np.array_split(Y, len(Y))):
        many.fit_many(xb, yb)

    for i in X:
        assert math.isclose(one.weights[i], many.weights[i])


def test_shuffle_columns():
    """Checks that fit_many works identically whether columns are shuffled or not."""

    X = pd.read_csv(datasets.TrumpApproval().path)
    Y = X.pop('five_thirty_eight')

    normal = linear_model.LinearRegression()
    for xb, yb in zip(np.array_split(X, 10), np.array_split(Y, 10)):
        normal.fit_many(xb, yb)

    shuffled = linear_model.LinearRegression()
    for xb, yb in zip(np.array_split(X, 10), np.array_split(Y, 10)):
        cols = np.random.permutation(X.columns)
        shuffled.fit_many(xb[cols], yb)

    for i in X:
        assert math.isclose(normal.weights[i], shuffled.weights[i])


def test_add_remove_columns():
    """Checks that no exceptions are raised whenever columns are dropped and/or added."""

    X = pd.read_csv(datasets.TrumpApproval().path)
    Y = X.pop('five_thirty_eight')

    lin_reg = linear_model.LinearRegression()
    for xb, yb in zip(np.array_split(X, 10), np.array_split(Y, 10)):
        # Pick half of the columns at random
        cols = np.random.choice(X.columns, len(X.columns) // 2, replace=False)
        lin_reg.fit_many(xb[cols], yb)



import collections
import functools
import typing

import creme.base

from . import base
from . import precision
from . import recall


__all__ = [
    'F1',
    'FBeta',
    'MacroF1',
    'MacroFBeta',
    'MicroF1',
    'MicroFBeta',
    'MultiFBeta',
    'WeightedF1',
    'WeightedFBeta'
]


class BaseFBeta:

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True


class FBeta(BaseFBeta, base.BinaryMetric):
    """Binary F-Beta score.

    The FBeta score is a weighted harmonic mean between precision and recall. The higher the
    `beta` value, the higher the recall will be taken into account. When `beta` equals 1,
    precision and recall and equivalently weighted, which results in the F1 score (see
    `creme.metrics.F1`).

    Parameters:
        beta: Weight of precision in the harmonic mean.

    Example:

        >>> from creme import metrics

        >>> y_true = [False, False, False, True, True, True]
        >>> y_pred = [False, False, True, True, False, False]

        >>> metric = metrics.FBeta(beta=2)
        >>> for y_t, y_p in zip(y_true, y_pred):
        ...     metric = metric.update(y_t, y_p)

        >>> metric
        FBeta: 0.357143

    """

    def __init__(self, beta: float):
        self.beta = beta

        self.precision = precision.Precision()
        """Precision score."""
        self.recall = recall.Recall()
        """Recall score."""

    def update(self, y_true, y_pred, sample_weight=1.):
        self.precision.update(y_true, y_pred, sample_weight)
        self.recall.update(y_true, y_pred, sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.precision.revert(y_true, y_pred, sample_weight)
        self.recall.revert(y_true, y_pred, sample_weight)
        return self

    def get(self):
        p = self.precision.get()
        r = self.recall.get()
        b2 = self.beta ** 2
        try:
            return (1 + b2) * p * r / (b2 * p + r)
        except ZeroDivisionError:
            return 0.


class MacroFBeta(BaseFBeta, base.MultiClassMetric):
    """Macro-average F-Beta score.

    This works by computing the F-Beta score per class, and then performs a global average.

    Parameters:
        beta (float): Weight of precision in harmonic mean.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.MacroFBeta(beta=.8)

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        MacroFBeta: 1.
        MacroFBeta: 0.310606
        MacroFBeta: 0.540404
        MacroFBeta: 0.540404
        MacroFBeta: 0.485982

    """

    def __init__(self, beta):
        self.fbetas = collections.defaultdict(functools.partial(FBeta, beta=beta))
        """The F-Beta score of each label."""
        self._class_counts = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self._class_counts.update([y_true, y_pred])

        for c in self._class_counts:
            self.fbetas[c].update(y_true == c, y_pred == c, sample_weight)

        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self._class_counts.subtract([y_true, y_pred])

        for c in self._class_counts:
            self.fbetas[c].revert(y_true == c, y_pred == c, sample_weight)

        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return sum(self.fbetas[c].get() for c in relevant) / len(relevant)
        except ZeroDivisionError:
            return 0.


class MicroFBeta(BaseFBeta, base.MultiClassMetric):
    """Micro-average F-Beta score.

    This computes the F-Beta score by merging all the predictions and true labels, and then
    computes a global F-Beta score.

    Parameters:
        beta: Weight of precision in the harmonic mean.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 0]
        >>> y_pred = [0, 1, 1, 2, 1]

        >>> metric = metrics.MicroFBeta(beta=2)
        >>> for y_t, y_p in zip(y_true, y_pred):
        ...     metric = metric.update(y_t, y_p)

        >>> metric
        MicroFBeta: 0.6

    References:
        1. [Why are precision, recall and F1 score equal when using micro averaging in a multi-class problem?](https://simonhessner.de/why-are-precision-recall-and-f1-score-equal-when-using-micro-averaging-in-a-multi-class-problem/)

    """

    def __init__(self, beta: float):
        self.beta = beta
        self.precision = precision.MicroPrecision()
        """Precision score."""
        self.recall = recall.MicroRecall()
        """Recall score."""

    def update(self, y_true, y_pred, sample_weight=1.):
        self.precision.update(y_true, y_pred, sample_weight)
        self.recall.update(y_true, y_pred, sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.precision.revert(y_true, y_pred, sample_weight)
        self.recall.revert(y_true, y_pred, sample_weight)
        return self

    def get(self):
        p = self.precision.get()
        r = self.recall.get()
        b2 = self.beta ** 2
        try:
            return (1 + b2) * p * r / (b2 * p + r)
        except ZeroDivisionError:
            return 0.


class WeightedFBeta(BaseFBeta, base.MultiClassMetric):
    """Weighted-average F-Beta score.

    This works by computing the F-Beta score per class, and then performs a global weighted average
    according to the support of each class.

    Parameters:
        beta: Weight of precision in the harmonic mean.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.WeightedFBeta(beta=0.8)

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        WeightedFBeta: 1.
        WeightedFBeta: 0.310606
        WeightedFBeta: 0.540404
        WeightedFBeta: 0.655303
        WeightedFBeta: 0.626283

    """

    def __init__(self, beta: float):
        self.fbetas: typing.DefaultDict[creme.base.typing.ClfTarget, FBeta] = collections.defaultdict(functools.partial(FBeta, beta=beta))
        self._support: typing.Counter[creme.base.typing.ClfTarget] = collections.Counter()
        self._class_counts: typing.Counter[creme.base.typing.ClfTarget] = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self._class_counts.update([y_true, y_pred])
        self._support.update({y_true: sample_weight})

        for c in self._class_counts:
            self.fbetas[c].update(y_true == c, y_pred == c, sample_weight)

        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self._class_counts.subtract([y_true, y_pred])
        self._support.subtract({y_true: sample_weight})

        for c in self._class_counts:
            self.fbetas[c].revert(y_true == c, y_pred == c, sample_weight)

        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return (
                sum(self.fbetas[c].get() * self._support[c] for c in relevant) /
                sum(self._support[c] for c in relevant)
            )
        except ZeroDivisionError:
            return 0.


class MultiFBeta(BaseFBeta, base.MultiClassMetric):
    """Multi-class F-Beta score with different betas per class.

    The multiclass F-Beta score is the arithmetic average of the binary F-Beta scores of each class.
    The mean can be weighted by providing class weights.

    Parameters:
        beta (dict): Weight of precision in harmonic mean. For different beta values per class,
            provide a dict with class labels as keys and beta values as values.
        weights: Class weights. If not provided then uniform weights will be used

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.MultiFBeta(
        ...     betas={0: 0.25, 1: 1, 2: 4},
        ...     weights={0: 1, 1: 1, 2: 2}
        ... )

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        MultiFBeta: 1.
        MultiFBeta: 0.257576
        MultiFBeta: 0.628788
        MultiFBeta: 0.628788
        MultiFBeta: 0.468788

    """

    def __init__(self, betas: typing.Dict[creme.base.typing.ClfTarget, float],
                 weights: typing.Dict[creme.base.typing.ClfTarget, float] = None):
        self.betas = betas
        self.fbetas: typing.Dict[creme.base.typing.ClfTarget, FBeta] = dict()
        self._class_counts: typing.Counter[creme.base.typing.ClfTarget] = collections.Counter()
        self.weights = (
            weights
            if weights is not None
            else collections.defaultdict(functools.partial(int, 1))
        )

    def update(self, y_true, y_pred, sample_weight=1.):
        self._class_counts.update([y_true, y_pred])

        for c in self._class_counts:
            try:
                fb = self.fbetas[c]
            except KeyError:
                fb = self.fbetas[c] = FBeta(beta=self.betas[c])

            fb.update(y_true=y_true == c, y_pred=y_pred == c, sample_weight=sample_weight)

        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self._class_counts.subtract([y_true, y_pred])

        for c in self._class_counts:
            try:
                fb = self.fbetas[c]
            except KeyError:
                fb = self.fbetas[c] = FBeta(beta=self.betas[c])

            fb.revert(y_true=y_true == c, y_pred=y_pred == c, sample_weight=sample_weight)

        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return (
                sum(self.fbetas[c].get() * self.weights[c] for c in relevant) /
                sum(self.weights[c] for c in relevant)
            )
        except ZeroDivisionError:
            return 0.


class F1(FBeta):
    """Binary F1 score.

    Example:

        >>> from creme import metrics

        >>> y_true = [False, False, False, True, True, True]
        >>> y_pred = [False, False, True, True, False, False]

        >>> metric = metrics.F1()
        >>> for y_t, y_p in zip(y_true, y_pred):
        ...     metric = metric.update(y_t, y_p)

        >>> metric
        F1: 0.4

    """

    def __init__(self):
        super().__init__(beta=1.)


class MacroF1(MacroFBeta):
    """Macro-average F1 score.

    This works by computing the F1 score per class, and then performs a global average.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.MacroF1()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        MacroF1: 1.
        MacroF1: 0.333333
        MacroF1: 0.555556
        MacroF1: 0.555556
        MacroF1: 0.488889

    """

    def __init__(self):
        super().__init__(beta=1.)


class MicroF1(MicroFBeta):
    """Micro-average F1 score.

    This computes the F1 score by merging all the predictions and true labels, and then computes a
    global F1 score.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 0]
        >>> y_pred = [0, 1, 1, 2, 1]

        >>> metric = metrics.MicroF1()
        >>> for y_t, y_p in zip(y_true, y_pred):
        ...     metric = metric.update(y_t, y_p)

        >>> metric
        MicroF1: 0.6

    References:
        1. [Why are precision, recall and F1 score equal when using micro averaging in a multi-class problem?](https://simonhessner.de/why-are-precision-recall-and-f1-score-equal-when-using-micro-averaging-in-a-multi-class-problem/)

    """

    def __init__(self):
        super().__init__(beta=1.)


class WeightedF1(WeightedFBeta):
    """Weighted-average F1 score.

    This works by computing the F1 score per class, and then performs a global weighted average by
    using the support of each class.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.WeightedF1()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        WeightedF1: 1.
        WeightedF1: 0.333333
        WeightedF1: 0.555556
        WeightedF1: 0.666667
        WeightedF1: 0.613333

    """

    def __init__(self):
        super().__init__(beta=1.)



import collections
import math
import random

from creme import base


class RBFSampler(base.Transformer):
    """Extracts random features which approximate an RBF kernel.

    This is a powerful way to give non-linear capacity to linear classifiers. In the litterature
    this is also called "random Fourier features".

    Parameters:
        gamma: RBF kernel parameter in `(-gamma * x^2)`.
        n_components: Number of samples per original feature. Equals the dimensionality of the
            computed feature space.
        seed: Random number seed.

    Example:

        >>> from creme import linear_model
        >>> from creme import optim
        >>> from creme import preprocessing
        >>> from creme import stream

        >>> # XOR function
        >>> X = [[0, 0], [1, 1], [1, 0], [0, 1]]
        >>> Y = [0, 0, 1, 1]

        >>> model = linear_model.LogisticRegression(optimizer=optim.SGD(0.1))

        >>> for x, y in stream.iter_array(X, Y):
        ...     model = model.fit_one(x, y)
        ...     y_pred = model.predict_one(x)
        ...     print(y, int(y_pred))
        0 0
        0 0
        1 0
        1 1

        >>> model = (
        ...     preprocessing.RBFSampler(seed=3) |
        ...     linear_model.LogisticRegression(optimizer=optim.SGD(0.1))
        ... )

        >>> for x, y in stream.iter_array(X, Y):
        ...     model = model.fit_one(x, y)
        ...     y_pred = model.predict_one(x)
        ...     print(y, int(y_pred))
        0 0
        0 0
        1 1
        1 1

    References:
        1. [Rahimi, A. and Recht, B., 2008. Random features for large-scale kernel machines. In Advances in neural information processing systems (pp. 1177-1184](https://people.eecs.berkeley.edu/~brecht/papers/07.rah.rec.nips.pdf)

    """

    def __init__(self, gamma=1., n_components=100, seed=None):
        self.gamma = gamma
        self.n_components = n_components
        self.seed = seed
        self.rng = random.Random(seed)
        self.weights = collections.defaultdict(self._random_weights)
        self.offsets = [random.uniform(0, 2 * math.pi)
                        for _ in range(n_components)]

    def _random_weights(self):
        return [
            math.sqrt(2 * self.gamma) * self.rng.gauss(mu=0, sigma=1)
            for _ in range(self.n_components)
        ]

    def transform_one(self, x, y=None):
        return {
            (i, j): math.cos(xi * wj + self.offsets[j])
            for i, xi in x.items()
            for j, wj in enumerate(self.weights[i])
        }



import abc
import collections
import numbers

from .. import optim
from .. import stats
from .. import utils


class BaseFM:
    """Factorization Machines base class.

    Parameters:
        n_factors (int): Dimensionality of the factorization or number of latent factors.
        weight_optimizer (optim.Optimizer): The sequential optimizer used for updating the feature
            weights. Note that the intercept is handled separately.
        latent_optimizer (optim.Optimizer): The sequential optimizer used for updating the latent
            factors.
        loss (optim.Loss): The loss function to optimize for.
        sample_normalization (bool): Whether to divide each element of ``x`` by ``x`` L2-norm.
            Defaults to False.
        l1_weight (float): Amount of L1 regularization used to push weights towards 0.
        l2_weight (float): Amount of L2 regularization used to push weights towards 0.
        l1_latent (float): Amount of L1 regularization used to push latent weights towards 0.
        l2_latent (float): Amount of L2 regularization used to push latent weights towards 0.
        intercept (float or `stats.Univariate` instance): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        weight_initializer (optim.initializers.Initializer): Weights initialization scheme. Defaults
            to ``optim.initializers.Zeros()``.
        latent_initializer (optim.initializers.Initializer): Latent factors initialization scheme.
            Defaults to
            ``optim.initializers.Normal(mu=.0, sigma=.1, random_state=self.random_state)``.
        clip_gradient (float): Clips the absolute value of each gradient value.
        seed (int): Randomization seed used for reproducibility.

    Attributes:
        weights (collections.defaultdict): The current weights assigned to the features.
        latents (collections.defaultdict): The current latent weights assigned to the features.

    """

    def __init__(self, n_factors, weight_optimizer, latent_optimizer, loss, sample_normalization,
                 l1_weight, l2_weight, l1_latent, l2_latent, intercept, intercept_lr,
                 weight_initializer, latent_initializer, clip_gradient, seed):
        self.n_factors = n_factors
        self.weight_optimizer = optim.SGD(0.01) if weight_optimizer is None else weight_optimizer
        self.latent_optimizer = optim.SGD(0.01) if latent_optimizer is None else latent_optimizer
        self.loss = loss
        self.sample_normalization = sample_normalization
        self.l1_weight = l1_weight
        self.l2_weight = l2_weight
        self.l1_latent = l1_latent
        self.l2_latent = l2_latent
        self.intercept = intercept

        self.intercept_lr = (
            optim.schedulers.Constant(intercept_lr)
            if isinstance(intercept_lr, numbers.Number) else
            intercept_lr
        )

        if weight_initializer is None:
            weight_initializer = optim.initializers.Zeros()
        self.weight_initializer = weight_initializer
        self.weights = collections.defaultdict(weight_initializer)

        if latent_initializer is None:
            latent_initializer = optim.initializers.Normal(sigma=.1, seed=seed)
        self.latent_initializer = latent_initializer
        self.latents = self._init_latents()

        self.clip_gradient = clip_gradient
        self.seed = seed

    @abc.abstractmethod
    def _init_latents(self) -> collections.defaultdict:
        """Initializes latent weights dict."""

    def fit_one(self, x, y, sample_weight=1.):
        x = self._ohe_cat_features(x)

        if self.sample_normalization:
            x_l2_norm = sum((xj ** 2 for xj in x.values())) ** 0.5
            x = {j: xj / x_l2_norm for j, xj in x.items()}

        return self._fit_one(x, y, sample_weight=sample_weight)

    def _ohe_cat_features(self, x):
        """One hot encodes string features considering them as categorical."""
        return dict((f'{j}_{xj}', 1) if isinstance(xj, str) else (j, xj) for j, xj in x.items())

    def _fit_one(self, x, y, sample_weight=1.):

        # Update the intercept if statistic before calculating the gradient
        if isinstance(self.intercept, stats.Univariate):
            self.intercept.update(y)

        # Calculate the gradient of the loss with respect to the raw output
        g_loss = self.loss.gradient(y_true=y, y_pred=self._raw_dot(x))

        # Clamp the gradient to avoid numerical instability
        g_loss = utils.math.clamp(g_loss, minimum=-self.clip_gradient, maximum=self.clip_gradient)

        # Apply the sample weight
        g_loss *= sample_weight

        # Update the intercept if not statistic
        if not isinstance(self.intercept, stats.Univariate):
            w0_lr = self.intercept_lr.get(self.weight_optimizer.n_iterations)
            self.intercept -= w0_lr * g_loss

        # Update the weights
        weights_gradient = self._calculate_weights_gradients(x, g_loss)
        self.weights = self.weight_optimizer.update_after_pred(w=self.weights, g=weights_gradient)

        # Update the latent weights
        self._update_latents(x, g_loss)

    def _raw_dot(self, x):

        # Start with the intercept
        intercept = self.intercept
        y_pred = intercept.get() if isinstance(intercept, stats.Univariate) else intercept

        # Add the unary interactions
        y_pred += utils.math.dot(x, self.weights)

        # Add greater than unary interactions
        y_pred += self._calculate_interactions(x)

        return y_pred

    def _field(self, j):
        """Infers feature field name."""
        return j.split('_')[0]

    @abc.abstractmethod
    def _calculate_interactions(self, x: dict) -> float:
        """Calculates greater than unary interactions."""

    @abc.abstractmethod
    def _calculate_weights_gradients(self, x: dict, g_loss: float) -> dict:
        """Calculates weights gradient."""

    @abc.abstractmethod
    def _update_latents(self, x: dict, g_loss: float):
        """Updates latent weights."""



import pytest
import torch

from .. import base


__all__ = [
    'PyTorch2CremeRegressor'
]


class PyTorch2CremeBase:

    def __init__(self, net, loss_fn, optimizer, batch_size, x_tensor, y_tensor):
        self.net = net
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.batch_size = batch_size
        self.x_tensor = x_tensor
        self.y_tensor = y_tensor
        self._x_batch = [None] * batch_size
        self._y_batch = [None] * batch_size
        self._batch_i = 0

    def fit_one(self, x, y):

        self._x_batch[self._batch_i] = list(x.values())
        self._y_batch[self._batch_i] = [y]

        self._batch_i += 1

        if self._batch_i == self.batch_size:

            x = self.x_tensor(self._x_batch)
            y = self.y_tensor(self._y_batch)

            self.optimizer.zero_grad()
            y_pred = self.net(x)
            loss = self.loss_fn(y_pred, y)
            loss.backward()
            self.optimizer.step()
            self._batch_i = 0

        return self


class PyTorch2CremeRegressor(PyTorch2CremeBase, base.Regressor):
    """PyTorch to `creme` regressor adapter.

    Parameters:
        net
        loss_fn
        optimizer
        batch_size

    Example:

        >>> from creme import compat
        >>> from creme import datasets
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import preprocessing
        >>> import torch
        >>> from torch import nn
        >>> from torch import optim

        >>> _ = torch.manual_seed(0)

        >>> X_y = datasets.TrumpApproval()

        >>> n_features = 6
        >>> net = nn.Sequential(
        ...     nn.Linear(n_features, 3),
        ...     nn.Linear(3, 1)
        ... )

        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     compat.PyTorch2CremeRegressor(
        ...         net=net,
        ...         loss_fn=nn.MSELoss(),
        ...         optimizer=optim.SGD(net.parameters(), lr=1e-3),
        ...         batch_size=2
        ...     )
        ... )
        >>> metric = metrics.MAE()

        >>> model_selection.progressive_val_score(X_y, model, metric).get()
        2.78258

    """

    def __init__(self, net: torch.nn.Sequential, loss_fn: torch.nn.modules.loss._Loss,
                 optimizer: torch.optim.Optimizer, batch_size=1):
        super().__init__(
            net=net,
            loss_fn=loss_fn,
            optimizer=optimizer,
            batch_size=batch_size,
            x_tensor=torch.Tensor,
            y_tensor=torch.Tensor
        )

    def predict_one(self, x):
        x = self.x_tensor(list(x.values()))
        return self.net(x).item()



from creme import base
from creme import utils


__all__ = ['PredClipper']


class PredClipper(base.Regressor, base.Wrapper):
    """Clips the target after predicting.

    Parameters:
        regressor: regressor model applied for training.
        y_min: minimum value.
        y_max: maximum value.

    Example:

        >>> from creme import linear_model
        >>> from creme import meta

        >>> X_y = (
        ...     ({'a': 2, 'b': 4}, 80),
        ...     ({'a': 3, 'b': 5}, 100),
        ...     ({'a': 4, 'b': 6}, 120)
        ... )

        >>> model = meta.PredClipper(
        ...     regressor=linear_model.LinearRegression(),
        ...     y_min=0,
        ...     y_max=200
        ... )

        >>> for x, y in X_y:
        ...     _ = model.fit_one(x, y)

        >>> model.predict_one({'a': -100, 'b': -200})
        0

        >>> model.predict_one({'a': 50, 'b': 60})
        200

    """

    def __init__(self, regressor: base.Regressor, y_min: float, y_max: float):
        self.regressor = regressor
        self.y_min = y_min
        self.y_max = y_max

    @property
    def _wrapped_model(self):
        return self.regressor

    @property
    def _labelloc(self):
        return 'b'  # for bottom

    def fit_one(self, x, y):
        self.regressor.fit_one(x, y)
        return self

    def predict_one(self, x):
        y_pred = self.regressor.predict_one(x)
        return utils.math.clamp(y_pred, self.y_min, self.y_max)



import collections
import math

from . import base


class Entropy(base.Univariate):
    """Running entropy.

    Parameters:
        alpha: Fading factor.
        eps: Small value that will be added to the denominator to avoid division by zero.

    Attributes:
        entropy (float): The running entropy.
        n (int): The current number of observations.
        counter (collections.Counter): Count the number of times the values have occurred

    Example:

        >>> import math
        >>> import random
        >>> import numpy as np
        >>> from scipy.stats import entropy
        >>> from creme import stats

        >>> def entropy_list(labels, base=None):
        ...   value,counts = np.unique(labels, return_counts=True)
        ...   return entropy(counts, base=base)

        >>> SEED = 42 * 1337
        >>> random.seed(SEED)

        >>> entro = stats.Entropy(alpha=1)

        >>> list_animal = []
        >>> for animal, num_val in zip(['cat', 'dog', 'bird'],[301, 401, 601]):
        ...     list_animal += [animal for i in range(num_val)]
        >>> random.shuffle(list_animal)

        >>> for animal in list_animal:
        ...     _ = entro.update(animal)

        >>> print(f'{entro.get():.6f}')
        1.058093
        >>> print(f'{entropy_list(list_animal):.6f}')
        1.058093


    References:
        1. [Sovdat, B., 2014. Updating Formulas and Algorithms for Computing Entropy and Gini Index from Time-Changing Data Streams. arXiv preprint arXiv:1403.6348.](https://arxiv.org/pdf/1403.6348.pdf)

    """

    def __init__(self, alpha=1, eps=1e-8):

        if 0 < alpha <= 1:
            self.alpha = alpha
        else:
            raise ValueError('alpha must be between 0 excluded and 1')
        self.eps = eps
        self.entropy = 0
        self.n = 0
        self.counter = collections.Counter()

    @property
    def name(self):
        return 'entropy'

    def update(self, x):

        cx = self.counter.get(x, 0)
        n = self.n
        eps = self.eps
        alpha = self.alpha

        entropy = self.entropy
        entropy = (n + eps) / (n + 1) * (alpha * entropy - math.log((n + eps) / (n + 1)))
        entropy -= (cx + 1) / (n + 1) * math.log((cx + 1) / (n + 1))
        entropy += (cx + eps) / (n + 1) * math.log((cx + eps) / (n + 1))
        self.entropy = entropy

        self.n += 1
        self.counter.update([x])

        return self

    def get(self):
        return self.entropy



import collections
import typing

import creme.base

from . import base


class MiniBatcher(base.Optimizer):
    """Wrapper class for doing mini-batch gradient descent.

    This will accumulate successive gradients until `batch_size` have been observed. Once the
    current mini-batch is full, the underlying optimizer will update the weights using the average
    gradient. This can significantly improve the accuracy of a model, especially early on when
    gradients tend to go wild.

    Parameters:
        optimizer
        batch_size: The desired batch size.

    Attributes:
        gradient (collections.defaultdict): The current accumulated mini-batch.
        current_size (int): The current size of the mini-batch.

    Example:

        >>> from creme import datasets
        >>> from creme import linear_model
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing

        >>> X_y = datasets.Phishing()
        >>> optimizer = optim.MiniBatcher(optim.SGD(0.1), 4)
        >>> model = (
        ...     preprocessing.StandardScaler() |
        ...     linear_model.LogisticRegression(optimizer)
        ... )
        >>> metric = metrics.F1()

        >>> model_selection.progressive_val_score(X_y, model, metric)
        F1: 0.88172

    """

    def __init__(self, optimizer: base.Optimizer, batch_size: int):
        self.optimizer = optimizer
        self.batch_size = batch_size
        self.gradient: typing.DefaultDict[creme.base.typing.FeatureName, float] = collections.defaultdict(float)
        self.current_size = 0
        self.n_iterations = 0

    def update_before_pred(self, w):
        return self.optimizer.update_before_pred(w)

    def _update_after_pred(self, w, g):

        # Accumulate the gradient
        for i, gi in g.items():
            self.gradient[i] += gi
        self.current_size += 1

        if self.current_size == self.batch_size:

            # Update the weights with the average gradient
            avg_g = {i: gi / self.batch_size for i, gi in self.gradient.items()}
            w = self.optimizer.update_after_pred(w, avg_g)

            # Reset the gradient
            self.gradient = collections.defaultdict(float)
            self.current_size = 0

        return w



from . import var


class SEM(var.Var):
    """Running standard error of the mean using Welford's algorithm.

    Parameters:
        ddof: Delta Degrees of Freedom. The divisor used in calculations is `n - ddof`, where `n`
            is the number of seen elements.

    Attributes:
        n (int): Number of observations.

    Example:

        >>> import creme.stats

        >>> X = [3, 5, 4, 7, 10, 12]

        >>> sem = creme.stats.SEM()
        >>> for x in X:
        ...     print(sem.update(x).get())
        0.0
        1.0
        0.577350
        0.853912
        1.240967
        1.447219

    References:
        1. [Wikipedia article on algorithms for calculating variance](https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Covariance)

    """

    def get(self):
        return (super().get() / self.mean.n) ** 0.5


class RollingSEM(var.RollingVar):
    """Running standard error of the mean over a window.

    Parameters:
        window_size: Size of the rolling window.
        ddof: Delta Degrees of Freedom for the variance.

    Example:

        >>> import creme

        >>> X = [1, 4, 2, -4, -8, 0]

        >>> rolling_sem = creme.stats.RollingSEM(ddof=1, window_size=2)
        >>> for x in X:
        ...     print(rolling_sem.update(x).get())
        0.0
        1.5
        1.0
        3.0
        2.0
        4.0

        >>> rolling_sem = creme.stats.RollingSEM(ddof=1, window_size=3)
        >>> for x in X:
        ...     print(rolling_sem.update(x).get())
        0.0
        1.5
        0.881917
        2.403700
        2.905932
        2.309401

    """

    def get(self):
        return (super().get() / len(self.rolling_mean)) ** 0.5



import abc
import typing

from creme import base

from . import estimator


class MultiOutputEstimator(estimator.Estimator):
    """A multi-output estimator."""


class MultiOutputClassifier(MultiOutputEstimator):
    """A multi-output classifier."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: typing.Dict[typing.Union[str, int], base.typing.ClfTarget]) -> 'MultiOutputClassifier':
        """Fits to a set of features `x` and a set of labels `y`.

        Parameters:
            x: A dictionary of features.
            y: A dictionary with the labels of each output.

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_proba_one(self, x: dict) -> typing.Dict[typing.Union[str, int], typing.Dict[base.typing.ClfTarget, float]]:
        """Given a set of features `x`, predicts a the probability of each label for each output.

        Parameters:
            x: A dictionary of features.

        Returns:
            A nested dictionary which contains the output probabilities of each output.

        """

    def predict_one(self, x: dict) -> typing.Dict[typing.Union[str, int], base.typing.ClfTarget]:
        """Given a set of features `x`, predicts a label for each output.

        Parameters:
            x: A dictionary of features.

        Returns:
            A dictionary with the predicted label for each output.

        """
        y_pred = self.predict_proba_one(x)
        return {
            c: max(y_pred[c], key=y_pred[c].get)
            for c in y_pred
        }


class MultiOutputRegressor(MultiOutputEstimator):
    """A multi-output regressor."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: typing.Dict[typing.Union[str, int], base.typing.RegTarget]) -> 'MultiOutputRegressor':
        """Fits to a set of features `x` and a set of outputs `y`.

        Parameters:
            x: A dictionary of features.
            y: A dictionary with the target of each output.

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_one(self, x: dict) -> typing.Dict[typing.Union[str, int], base.typing.RegTarget]:
        """Given a set of features `x`, predicts a target for each output.

        Parameters:
            x: A dictionary of features.

        Returns:
            A dictionary with the predicted outcomes of each output.

        """



import abc


class Distribution(abc.ABC):

    @abc.abstractmethod
    def update(self, x):
        """Updates the parameters of the distribution given a new observation."""

    @abc.abstractproperty
    def n_samples(self):
        """The number of observed samples."""

    def __repr__(self):
        return str(self)


class DiscreteDistribution(Distribution):
    """A probability distribution for discrete values."""

    @abc.abstractmethod
    def pmf(self, x):
        """Probability mass function."""


class ContinuousDistribution(Distribution):
    """A probability distribution for continuous values."""

    @abc.abstractproperty
    def mode(self):
        """Most likely value."""

    @abc.abstractmethod
    def pdf(self, x):
        """Probability density function, i.e. P(x <= X < x+dx) / dx."""

    @abc.abstractmethod
    def cdf(self, x):
        """Cumulative density function, i.e. P(X <= x)."""



import abc
import numbers
import typing

from creme import base

from .predictor import Predictor


class Regressor(Predictor):
    """A regressor."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: base.typing.RegTarget) -> 'Regressor':
        """Fits to a set of features ``x`` and a real-valued target ``y``.

        Parameters:
            x: A dictionary of features.
            y: A numeric target.

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_one(self, x: dict) -> base.typing.RegTarget:
        """Predicts the target value of a set of features ``x``.

        Parameters:
            x: A dictionary of features.

        Returns:
            The prediction.

        """



import collections

from .. import stats

from . import base


__all__ = [
    'MacroPrecision',
    'MicroPrecision',
    'Precision',
    'WeightedPrecision'
]


class BasePrecision:

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True


class Precision(BasePrecision, base.BinaryMetric):
    """Binary precision score.

    Example:

        >>> from creme import metrics

        >>> y_true = [True, False, True, True, True]
        >>> y_pred = [True, True, False, True, True]

        >>> metric = metrics.Precision()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        Precision: 1.
        Precision: 0.5
        Precision: 0.5
        Precision: 0.666667
        Precision: 0.75

    """

    def __init__(self):
        self._mean = stats.Mean()

    def update(self, y_true, y_pred, sample_weight=1.):
        if y_pred:
            self._mean.update(x=y_true == y_pred, w=sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        if y_pred:
            self._mean.revert(x=y_true == y_pred, w=sample_weight)
        return self

    def get(self):
        return self._mean.get()


class MacroPrecision(BasePrecision, base.MultiClassMetric):
    """Macro-average precision score.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.MacroPrecision()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        MacroPrecision: 1.
        MacroPrecision: 0.25
        MacroPrecision: 0.5
        MacroPrecision: 0.5
        MacroPrecision: 0.5

    """

    def __init__(self):
        self.precisions = collections.defaultdict(Precision)
        self._class_counts = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self.precisions[y_pred].update(y_true == y_pred, True, sample_weight)
        self._class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.precisions[y_pred].revert(y_true == y_pred, True, sample_weight)
        self._class_counts.subtract([y_true, y_pred])
        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return sum(self.precisions[c].get() for c in relevant) / len(relevant)
        except ZeroDivisionError:
            return 0.


class MicroPrecision(BasePrecision, base.MeanMetric, base.MultiClassMetric):
    """Micro-average precision score.

    The micro-average precision score is exactly equivalent to the micro-average recall as well as
    the micro-average F1 score.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.MicroPrecision()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        MicroPrecision: 1.
        MicroPrecision: 0.5
        MicroPrecision: 0.666667
        MicroPrecision: 0.75
        MicroPrecision: 0.6

    References:
        1. [Why are precision, recall and F1 score equal when using micro averaging in a multi-class problem?](https://simonhessner.de/why-are-precision-recall-and-f1-score-equal-when-using-micro-averaging-in-a-multi-class-problem/)

    """

    def _eval(self, y_true, y_pred):
        return y_true == y_pred


class WeightedPrecision(BasePrecision, base.MultiClassMetric):
    """Weighted-average precision score.

    This uses the support of each label to compute an average score, whereas
    `metrics.MacroPrecision` ignores the support.

    Example:

        >>> from creme import metrics

        >>> y_true = [0, 1, 2, 2, 2]
        >>> y_pred = [0, 0, 2, 2, 1]

        >>> metric = metrics.WeightedPrecision()

        >>> for yt, yp in zip(y_true, y_pred):
        ...     print(metric.update(yt, yp))
        WeightedPrecision: 1.
        WeightedPrecision: 0.25
        WeightedPrecision: 0.5
        WeightedPrecision: 0.625
        WeightedPrecision: 0.7

    """

    def __init__(self):
        self.precisions = collections.defaultdict(Precision)
        self.support = collections.Counter()
        self._class_counts = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self.precisions[y_pred].update(y_true == y_pred, True, sample_weight)
        self.support.update({y_true: sample_weight})
        self._class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.precisions[y_pred].revert(y_true == y_pred, True, sample_weight)
        self.support.subtract({y_true: sample_weight})
        self._class_counts.subtract([y_true, y_pred])
        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return (
                sum(self.precisions[c].get() * self.support[c] for c in relevant) /
                sum(self.support[c] for c in relevant)
            )
        except ZeroDivisionError:
            return 0.
